"""Tests that ensure the `monitor` decorator takes its config from env."""

import os
import sys
from unittest import mock

import pytest


@pytest.mark.parametrize(
    ["env_vars", "error_message"],
    [
        ({}, "CRON_MON_SERVER_URL"),
        ({"CRON_MON_SERVER_URL": "http://localhost:8000"}, "CRON_MON_API_KEY"),
        ({"CRON_MON_API_KEY": "test"}, "CRON_MON_SERVER_URL"),
    ],
)
def test_monitor_env(env_vars: dict[str, str], error_message: str) -> None:
    """Test that the `monitor` decorator takes its config from env."""
    # when `cron_mon.monitor` is imported in other tests, it will be pre-loaded
    # with the default env vars. We need to remove it from `sys.modules` to
    # ensure it is re-imported with the new env vars in this test.
    sys.modules.pop("cron_mon.monitor", None)

    with mock.patch.dict(os.environ, env_vars, clear=True):
        with pytest.raises(KeyError, match=error_message):
            from cron_mon.monitor import monitor  # noqa: F401



"""Constants for tests."""

# Helpful constants for testing.
MONITOR_ID = "a04376e2-0fb5-4949-9744-7c5d0a50b411"
API_START_URL = (
    f"http://mock.cron-mon.io/api/v1/monitors/{MONITOR_ID}/jobs/start"
)
API_FINISH_URL = (
    f"http://mock.cron-mon.io/api/v1/monitors/{MONITOR_ID}/jobs/12345/finish"
)



"""Exceptions for the cron_mon package."""


class CronMonException(Exception):
    """Base exception for the cron_mon package."""


class CronMonAPIException(CronMonException):
    """Exception for API errors from the CronMon server."""


class InvalidAPIKey(CronMonAPIException):
    """Exception for an invalid API key."""



"""Decorator for monitoring function execution time via CronMon."""

import os
from functools import wraps
from typing import Any, Callable, Optional

from requests import ConnectionError, Session

from .exceptions import CronMonAPIException, InvalidAPIKey


class monitor:
    """A decorator class for monitoring function execution time via CronMon."""

    __SERVER_URL = os.environ["CRON_MON_SERVER_URL"]
    __API_KEY = os.environ["CRON_MON_API_KEY"]
    # TODO: Logger?

    def __init__(self, monitor_id: str) -> None:
        """Initialize the decorator with the monitor ID to use for monitoring.

        Args:
            monitor_id: The monitor ID to use for monitoring the function.
        """
        self.monitor_id = monitor_id

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap the function with monitoring logic.

        Args:
            func: The function to wrap with monitoring logic.
        """

        @wraps(func)
        def wrapper(*args: tuple, **kwargs: dict) -> Any:
            with Session() as session:
                job_id = self.__record_start(session)

                # Execute the decorated function and record the output and any
                # exceptions that occur.
                exc = None
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    exc = e

                output = exc or result
                self.__record_finish(
                    session,
                    job_id,
                    exc is None,
                    str(output) if output else None,
                )

                # Raise an exception if one occurred, since we don't want to
                # swallow exceptions, just record them. The underlying
                # application can decide if/ how to handle them.
                if exc:
                    raise exc

                return result

        return wrapper

    def __record_start(self, session: Session) -> str:
        """Record the start of a job with the CronMon server.

        Args:
            session: The session to use for making the API call.

        Returns:
            The job ID for the started job.
        """
        return self.__api_call(
            session, f"/monitors/{self.monitor_id}/jobs/start"
        )["job_id"]

    def __record_finish(
        self,
        session: Session,
        job_id: str,
        succeeded: bool,
        output: Optional[str],
    ) -> None:
        """Record the finish of a job with the CronMon server.

        Args:
            session: The session to use for making the API call.
            job_id: The job ID for the job to finish.
            succeeded: Whether the job succeeded or not.
            output: The output of the job.
        """
        self.__api_call(
            session,
            f"/monitors/{self.monitor_id}/jobs/{job_id}/finish",
            json={"succeeded": succeeded, "output": output},
        )

    def __api_call(
        self, session: Session, endpoint: str, json: Optional[dict] = None
    ) -> dict:
        """Make an API call to the CronMon server.

        Args:
            session: The session to use for making the API call.
            endpoint: The API endpoint to call.
            json: The JSON data to send with the API call.

        Returns:
            The JSON response from the API call.
        """
        try:
            response = session.post(
                url=f"{self.__SERVER_URL}/api/v1{endpoint}",
                headers={"X-API-Key": self.__API_KEY},
                json=json,
            )
        except ConnectionError as e:
            raise CronMonAPIException(
                f"Failed to connect to the CronMon API at {self.__SERVER_URL}"
            ) from e

        if not response.ok:
            message = response.json()["error"]["description"]
            if response.status_code == 401:
                raise InvalidAPIKey(message)
            raise CronMonAPIException(message)

        return response.json()["data"]



"""Basic tests for the `monitor` decorator on the 'happy path'."""

import pytest
import responses
from responses import matchers

from cron_mon import monitor
from tests.constants import API_FINISH_URL, API_START_URL, MONITOR_ID


@responses.activate
def test_monitoring() -> None:
    """Test that the `monitor` decorator works as expected."""
    responses.post(
        API_START_URL,
        json={"data": {"job_id": "12345"}},
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )
    responses.post(
        API_FINISH_URL,
        json={"data": {}},
        match=[
            matchers.header_matcher({"X-API-Key": "mock-api-key"}),
            matchers.json_params_matcher(
                {"succeeded": True, "output": "Hello, world!"}
            ),
        ],
    )

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        return "Hello, world!"

    cron_job()

    assert [c.request.url for c in responses.calls] == [
        API_START_URL,
        API_FINISH_URL,
    ]


@responses.activate
def test_monitor_failing_job() -> None:
    """Test that the `monitor` decorator handles failing jobs."""
    responses.post(
        API_START_URL,
        json={"data": {"job_id": "12345"}},
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )
    responses.post(
        API_FINISH_URL,
        json={"data": {}},
        match=[
            matchers.header_matcher({"X-API-Key": "mock-api-key"}),
            matchers.json_params_matcher(
                {"succeeded": False, "output": "Job failed"}
            ),
        ],
    )

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        raise ValueError("Job failed")

    # Any exceptions raised by the job should be propagated up to the caller.
    with pytest.raises(ValueError):
        cron_job()

    assert [c.request.url for c in responses.calls] == [
        API_START_URL,
        API_FINISH_URL,
    ]



"""CronMon integration package."""

from .exceptions import CronMonAPIException, CronMonException, InvalidAPIKey
from .monitor import monitor

__all__ = [
    "monitor",
    "CronMonException",
    "CronMonAPIException",
    "InvalidAPIKey",
]



"""Tests for the `monitor` decorator's error handling."""

import pytest
import responses
from responses import matchers

from cron_mon import CronMonAPIException, InvalidAPIKey, monitor
from tests.constants import API_FINISH_URL, API_START_URL, MONITOR_ID


def test_connection_errors() -> None:
    """Test that the `monitor` decorator handles connection errors."""

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        return "Hello, world!"

    with pytest.raises(CronMonAPIException):
        cron_job()


@responses.activate
def test_failed_start() -> None:
    """Test that the `monitor` decorator handles failed job starts."""
    responses.post(
        API_START_URL,
        status=500,
        json={
            "error": {
                "code": "500",
                "reason": "Internal Server Error",
                "description": "Failed to start job",
            }
        },
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        return "Hello, world!"

    with pytest.raises(CronMonAPIException) as exc:
        cron_job()

    assert str(exc.value) == "Failed to start job"


@responses.activate
def test_failed_finish() -> None:
    """Test that the `monitor` decorator handles failed job finishes."""
    responses.post(
        API_START_URL,
        json={"data": {"job_id": "12345"}},
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )
    responses.post(
        API_FINISH_URL,
        status=500,
        json={
            "error": {
                "code": "500",
                "reason": "Internal Server Error",
                "description": "Failed to finish job",
            }
        },
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        return "Hello, world!"

    with pytest.raises(CronMonAPIException) as exc:
        cron_job()

    assert str(exc.value) == "Failed to finish job"


@responses.activate
def test_invalid_api_key() -> None:
    """Test that the `monitor` decorator handles failed job finishes with output."""
    responses.post(
        API_START_URL,
        status=401,
        json={
            "error": {
                "code": 401,
                "reason": "Unauthorized",
                "description": "Unauthorized: Invalid API key",
            }
        },
        match=[matchers.header_matcher({"X-API-Key": "mock-api-key"})],
    )

    @monitor(MONITOR_ID)
    def cron_job() -> str:
        """A sample cron job."""
        return "Hello, world!"

    with pytest.raises(InvalidAPIKey) as exc:
        cron_job()

    assert str(exc.value) == "Unauthorized: Invalid API key"
